////////////////////////////////////////////////////////////////////// 
// lamp/Validator.cpp 
// (c) 2000-2007 Goncalo Abecasis
// 
// This file is distributed as part of the LAMP source code package   
// and may not be redistributed in any form, without prior written    
// permission from the author. Permission is granted for you to       
// modify this file for your own personal use, but modified versions  
// must retain this copyright notice and must not be distributed.     
// 
// Permission is granted for you to use this file to compile LAMP.    
// 
// All computer programs have bugs. Use this file at your own risk.   
// 
// Friday February 16, 2007
// 
 
#include "Validator.h"
#include "Kinship.h"

void Validator::TrimPedigree(Pedigree & ped, MarkerMap & map1, MarkerMap & map2)
   {
   IntArray markerIds;

   // Extract a list of markers
   for (int i = 0, markerId; i < map1.markers.Length(); i++)
      if ( (markerId = ped.LookupMarker(map1.markers[i])) >= 0)
         markerIds.Push(markerId);

   for (int i = 0, markerId; i < map2.markers.Length(); i++)
      if ( (markerId = ped.LookupMarker(map2.markers[i])) >= 0)
         markerIds.Push(markerId);

   // If there are no informative markers, print a warning message
   if (markerIds.Length() == 0)
      {
      printf("I don't know what markers to include in analysis!\n"
             "Please check that the framework and candidate map\n"
             "files are available and correctly formatted.\n\n");

      return;
      }

   // List informative individuals for fitting our retrospective likelihood
   IntArray informative(ped.count);

   // By default, we assume all individuals are uninformative
   informative.Set(false);

   // Individuals with genotype data are always informative
   for (int i = 0; i < ped.count; i++)
      for (int j = 0; j < markerIds.Length(); j++)
         if (ped[i].markers[markerIds[j]].isKnown())
            {
            informative[i] = true;
            break;
            }

   // Phenotyped relatives of individuals with genotype data are also informative
   for (int i = 0; i < ped.familyCount; i++)
      {
      bool informativeFamily = false;

      // First, check if we have some genotype data available for the family
      for (int j = ped.families[i]->first; j <= ped.families[i]->last; j++)
         if (informative[j])
            {
            informativeFamily = true;
            break;
            }

      if (!informativeFamily) break;

      // Then, check if we have phenotype data available for the untyped individuals
      for (int j = ped.families[i]->first; j <= ped.families[i]->last; j++)
         if (!informative[j])
            informative[j] = isPhenotyped(ped[j]);

      // TODO -- The above check is a little generous -- it would be best to
      //         ensure that phenotyped individuals have kinship > 1 with at
      //         least one genotyped individual.
      }

   // Finally, remove all individuals without genotype data
   // from the pedigree, as these will not contribute to our
   // likelihood test
   ped.Trim(false, informative);
   }

bool Validator::isPhenotyped(Person & p)
   {
   // In this simple version, we just check if affection status information
   // is available ...
   for (int k = 0; k < p.affectionCount; k++)
      if (p.affections[k])
         return true;

   return false;
   }

int Validator::EvaluateSample(Pedigree & ped, int affection)
   {
   // This variable encodes the phenotypic configuration for the last sibpair family seen ...
   int  lastSibPair = 0;

   // Track hard to classify families ...
   int  hardToClassify = 0;

   for (int i = 0; i < ped.familyCount; i++)
      {
      Family * fam = ped.families[i];

      // Check if we are dealing with a strange family we can't handle
      if (fam->founders > 2 && fam->generations == 2)
         {
         hardToClassify++;
         continue;
         }

      // Check if we are dealing with unrelated individuals or trios ...
      if (fam->founders + 1 >= fam->count)
         continue;

      // Nuclear families can contribute information as for "unrelateds",
      // "trios", "sibpairs" or more extended pedigrees depending on who
      // is phenotyped
      if (fam->isNuclear())
         {
         // Offspring phenotypes
         int kids[2], phenotypedKids = 0;

         // Count phenotyped offspring ...
         for (int j = fam->founders; j < fam->count; j++)
            if (ped[fam->path[j]].affections[affection])
               {
               if (phenotypedKids < 2)
                  kids[phenotypedKids] = ped[fam->path[j]].affections[affection];

               phenotypedKids++;
               }

         // If there are less than two phenotyped kids, we can't fit a linkage model
         if (phenotypedKids <= 1)
            continue;

         // If we have exactly two phenotyped kids, then we check whether the family
         // looks "different" than the last sibpair we saw ...
         if (phenotypedKids == 2)
            {
            // Parental phenotypes
            int parents[2];

            parents[0] = ped[fam->path[0]].affections[affection];
            parents[1] = ped[fam->path[1]].affections[affection];

            // Encode parental phenotypes
            int encoded = (parents[0] > parents[1]) ? parents[0] * 4 + parents[1] :
                                                      parents[1] * 4 + parents[0];

            // Encode offspring phenotypes
            encoded += (kids[0] > kids[1]) ? kids[0] * 64 + kids[1] * 16 :
                                             kids[1] * 64 + kids[0] * 16;

            // If all sibpair families look the same, we can only fit a
            // restricted linkage model
            if (lastSibPair == 0 || lastSibPair == encoded)
               {
               lastSibPair = encoded;
               continue;
               }
            }
         }

      // We only get here if there are different types of sibpairs in
      // the sample or if we encounter a more extended pedigree. In this
      // situation the sample allows a full linkage model to be fitted

      printf("The sample includes extended pedigrees and/or distinct sibpair configurations\n"
             "and should allow all model parameters to be estimated\n\n");
      return FULL_LINKAGE_INFORMATION;
      }

   // Sample includes some sibpairs, but they are all of the same type
   if (lastSibPair)
      {
      printf("This is a sibpair sample that does not allow all parameters to be estimated\n"
             "because only 2 degrees of freedom are available to model linkage. For linkage\n"
             "tests, the disease allele frequency will be fixed unless you specify an\n"
             "additive, recessive, dominant or multiplicative model.\n\n");
      return SOME_LINKAGE_INFORMATION;
      }

   // Sample consists only of trios and unrelated individuals
   printf("The sample includes only unrelated individuals and/or trios\n"
          "No information on genetic linkage is available\n\n");
   return NO_LINKAGE_INFORMATION;
   }

bool Validator::UnrelatedsOnly(Pedigree & ped)
   {
   Kinship kin;

   for (int i = 0; i < ped.familyCount; i++)
      if (ped.families[i]->count != ped.families[i]->founders)
         {
         kin.Setup(*(ped.families[i]));

         for (int j = ped.families[i]->first; j <= ped.families[i]->last; j++)
            if (isInformative(ped[j]))
               for (int k = j + 1; k < ped.families[i]->last; k++)
                  if (isInformative(ped[k]) && kin(ped[j], ped[k]) > 0)
                     return false;
         }

   return true;
   }

int Validator::FastFamilyBits(Pedigree & ped)
   {
   int bits = 0;

   for (int i = 0; i < ped.familyCount; i++)
      if (2 * ped.families[i]->nonFounders - ped.families[i]->founders > bits)
         bits = 2 * ped.families[i]->nonFounders - ped.families[i]->founders;

   return bits;
   }


int Validator::FamilyBits(Pedigree & ped)
   {
   IntArray informative;
   int bits = 0;

   for (int i = 0; i < ped.familyCount; i++)
      if (2 * ped.families[i]->nonFounders - ped.families[i]->founders > bits)
         {
         int founders = 0;
         int nonFounders = 0;

         informative.Dimension(ped.families[i]->count);
         informative.Zero();

         for (int j = ped.families[i]->count - 1; j >= 0; j--)
            {
            Person & p = ped[ped.families[i]->path[j]];

            if (informative[j] || isInformative(p))
               {
               if (p.isFounder())
                  founders++;
               else
                  {
                  nonFounders++;
                  informative[p.father->serial] = 1;
                  informative[p.mother->serial] = 1;
                  }
               }
            }

         if (2 * nonFounders - founders > bits)
            bits = 2 * nonFounders - founders;
         }

   return bits;
   }

bool Validator::isInformative(Person & p)
   {
   for (int i=0; i < p.affectionCount; i++)
      if (p.affections[i] > 0)
         return true;

   for (int i=0; i <= p.markerCount; i++) 
      if (p.markers[i].isKnown()) 
         return true;

   return false;
   }
 
